import 'package:adeptarch_base/src/pagination/helpers/grid_helpers.dart';
import 'package:flutter/material.dart';
//import 'package:flutter_pagewise/flutter_pagewise.dart';
import 'package:adeptarch_base/src/pagination/flutter_pagewise.dart'; //this implementation fixes a bug in resetting and fetching data too fast
import 'package:adeptarch_base/src/pagination/request_paginate.dart';
import 'package:adeptutils/adeptutils.dart';
import 'package:adeptarch_base/src/resource/resource.dart';
import 'dart:async';

typedef PageRequest = Future Function(int pageIndex);
typedef ErrorNotifier = void Function(
    BuildContext context, Resource resourceError);
typedef TextNoDataCallback = String Function();

class PaginateController<T> extends PagewiseLoadController<T> {
  String defaultTextNoData;
  TextNoDataCallback? textNoDataCallback;

  PaginateController({
    PageFuture<T>? pageControl,
    PageRequest? pageRequest,
    int pageSize = RequestPaginate.DEFAULT_SIZE,
    this.defaultTextNoData = "No data to show",
    this.textNoDataCallback,
  }) : super(
            pageFuture: pageControl ??
                (index) async {
                  try {
                    Object? response;
                    List<T> result = [];
                    if (pageRequest != null) {
                      response = await pageRequest(index);
                    }
                    if (response is Resource) {
                      result = _checkResource<T>(
                          response as Resource<List<T>?>,
                          textNoDataCallback != null
                              ? textNoDataCallback()
                              : defaultTextNoData);
                    } else if (response is List) {
                      result = response as List<T>;
                    }
                    if (index == 0 && result.isEmpty) {
                      throw Resource.error(
                          null,
                          textNoDataCallback != null
                              ? textNoDataCallback()
                              : defaultTextNoData);
                    }
                    return result;
                  } catch (e) {
                    return Future.error(e);
                  }
                },
            pageSize: pageSize);

  static List<T> _checkResource<T>(
      Resource<List<T>?> resource, String defaultTextNoData) {
    switch (resource.status) {
      case ResourceStatus.SUCCESS:
        if (resource.exists()) {
          return resource.data ?? [];
        } else {
          resource.status = ResourceStatus.ERROR;
          resource.message = defaultTextNoData;
          throw resource;
        }
      case ResourceStatus.TRANSIENT:
      case ResourceStatus.LOADING:
        return [];
      case ResourceStatus.ERROR:
      default:
        throw resource;
    }
  }

  static errorBuilder<T>({
    required BuildContext context,
    Object? error,
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
    PagewiseLoadController<T>? pageLoadController,
  }) {
    Resource resourceError = Resource.error(null, defaultTextError);
    try {
      if (error is Exception) {
        Exception exc = ExceptionUtils.parseException(error);
        String message = (getLocalizedMessageFromException != null
                ? getLocalizedMessageFromException(exc)
                : null) ??
            exc.toString();
        resourceError = Resource.error(null, message, exception: exc);
      } else if (error is String) {
        resourceError = Resource.error(null, error);
      }
      if (error is Resource) resourceError = error;

      if (TextUtils.isEmpty(resourceError.message)) {
        if (resourceError.exception != null) {
          Exception exc = ExceptionUtils.parseException(error);
          String message = (getLocalizedMessageFromException != null
                  ? getLocalizedMessageFromException(exc)
                  : null) ??
              exc.toString();
          resourceError.message = message;
        } else if (resourceError.code != null) {
          resourceError.message = (getLocalizedMessageFromHttpResultCode != null
              ? getLocalizedMessageFromHttpResultCode(resourceError.code)
              : null);
        }
      }

      if (TextUtils.isEmpty(resourceError.message)) {
        resourceError.message = defaultTextError;
      }
    } catch (e) {
      Exception exc = ExceptionUtils.parseException(error);
      String message = (getLocalizedMessageFromException != null
              ? getLocalizedMessageFromException(exc)
              : null) ??
          exc.toString();
      resourceError = Resource.error(null, message, exception: exc);
    }

    if (errorNotifier != null) {
      errorNotifier(context, resourceError);
    }
    if (errorViewBuilder != null) {
      return errorViewBuilder(
          context, resourceError.message, pageLoadController);
    }
    return Column(
      children: <Widget>[
        const Icon(
          Icons.error_outline,
          size: 32,
        ),
        const SizedBox(
          height: 4,
        ),
        Text(
          '${resourceError.message}',
          textAlign: TextAlign.center,
        ),
        const SizedBox(height: 8),
        FloatingActionButton(
            mini: true,
            onPressed: () {
              if (pageLoadController != null) {
                pageLoadController.retry();
              }
            },
            child: const Icon(
              Icons.refresh,
            )),
      ],
    );
  }
}

typedef LocalizedMessageFromException = String Function(dynamic exception);
typedef LocalizedMessageFromHttpResultCode = String Function(int? code);
typedef ErrorViewBuilder = Widget Function(BuildContext context,
    String? message, PagewiseLoadController? pageLoadController);

class PaginatedListView<T> extends PagewiseListView<T> {
  PaginatedListView({
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    super.key,
    super.padding,
    super.primary,
    super.addSemanticIndexes,
    super.semanticChildCount,
    super.shrinkWrap,
    super.controller,
    super.pageLoadController,
    super.itemExtent,
    super.addAutomaticKeepAlives,
    super.scrollDirection,
    super.addRepaintBoundaries,
    super.keyboardDismissBehavior,
    super.cacheExtent,
    ScrollPhysics super.physics = const AlwaysScrollableScrollPhysics(),
    super.reverse,
    super.pageSize,
    super.pageFuture,
    LoadingBuilder? loadingBuilder,
    super.retryBuilder,
    super.noItemsFoundBuilder,
    super.showRetry = false,
    required super.itemBuilder,
    super.itemDummyBuilder,
    super.itemDummyCount,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
    ErrorBuilder? errorBuilder,
  }) : super(
          loadingBuilder:
              loadingBuilder ?? (context) => const RefreshProgressIndicator(),
          errorBuilder: errorBuilder ??
              (context, error) => PaginateController.errorBuilder(
                    context: context,
                    error: error,
                    defaultTextError: defaultTextError,
                    getLocalizedMessageFromException:
                        getLocalizedMessageFromException,
                    getLocalizedMessageFromHttpResultCode:
                        getLocalizedMessageFromHttpResultCode,
                    errorNotifier: errorNotifier,
                    errorViewBuilder: errorViewBuilder,
                    pageLoadController: pageLoadController,
                  ),
        );
}

class PaginatedGridView<T> extends PagewiseGridView<T> {
  PaginatedGridView.count(
      {String defaultTextError = "Unknown error",
      LocalizedMessageFromException? getLocalizedMessageFromException,
      LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
      super.key,
      super.padding,
      required super.crossAxisCount,
      super.childAspectRatio,
      super.crossAxisSpacing,
      super.mainAxisSpacing,
      super.addSemanticIndexes,
      super.semanticChildCount,
      super.primary,
      super.shrinkWrap,
      super.controller,
      super.pageLoadController,
      super.addAutomaticKeepAlives,
      super.scrollDirection,
      super.addRepaintBoundaries,
      super.keyboardDismissBehavior,
      super.cacheExtent,
      ScrollPhysics super.physics = const AlwaysScrollableScrollPhysics(),
      super.reverse,
      super.pageSize,
      super.pageFuture,
      LoadingBuilder? loadingBuilder,
      super.retryBuilder,
      super.noItemsFoundBuilder,
      super.showRetry = false,
      required super.itemBuilder,
      super.itemDummyBuilder,
      super.itemDummyCount,
      ErrorNotifier? errorNotifier,
      ErrorViewBuilder? errorViewBuilder,
      ErrorBuilder? errorBuilder})
      : super.count(
          loadingBuilder:
              loadingBuilder ?? (context) => const RefreshProgressIndicator(),
          errorBuilder: errorBuilder ??
              (context, error) => PaginateController.errorBuilder(
                    context: context,
                    error: error,
                    defaultTextError: defaultTextError,
                    getLocalizedMessageFromException:
                        getLocalizedMessageFromException,
                    getLocalizedMessageFromHttpResultCode:
                        getLocalizedMessageFromHttpResultCode,
                    errorNotifier: errorNotifier,
                    errorViewBuilder: errorViewBuilder,
                    pageLoadController: pageLoadController,
                  ),
        );
}

class PaginatedSliverList<T> extends Pagewise<T> {
  static Widget withExtras<T>({
    Key? key,
    EdgeInsetsGeometry? padding,
    bool addSemanticIndexes = true,
    bool addAutomaticKeepAlives = true,
    bool addRepaintBoundaries = true,
    SemanticIndexCallback? semanticIndexCallback,
    int semanticIndexOffset = 0,
    PagewiseLoadController<T>? pageLoadController,
    int? pageSize,
    PageFuture<T>? pageFuture,
    LoadingBuilder? loadingBuilder,
    RetryBuilder? retryBuilder,
    NoItemsFoundBuilder? noItemsFoundBuilder,
    bool showRetry = false,
    required ItemBuilder<T> itemBuilder,
    ItemBuilder<T>? itemDummyBuilder,
    int? itemDummyCount,
    ErrorBuilder? errorBuilder,
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
  }) {
    var sliver = PaginatedSliverList<T>(
      key: key,
      addSemanticIndexes: addSemanticIndexes,
      addAutomaticKeepAlives: addAutomaticKeepAlives,
      addRepaintBoundaries: addRepaintBoundaries,
      semanticIndexCallback: semanticIndexCallback,
      semanticIndexOffset: semanticIndexOffset,
      pageLoadController: pageLoadController,
      pageSize: pageSize,
      pageFuture: pageFuture,
      loadingBuilder: loadingBuilder,
      retryBuilder: retryBuilder,
      noItemsFoundBuilder: noItemsFoundBuilder,
      showRetry: showRetry,
      itemBuilder: itemBuilder,
      itemDummyBuilder: itemDummyBuilder,
      itemDummyCount: itemDummyCount,
      errorBuilder: errorBuilder,
      defaultTextError: defaultTextError,
      getLocalizedMessageFromException: getLocalizedMessageFromException,
      getLocalizedMessageFromHttpResultCode:
          getLocalizedMessageFromHttpResultCode,
      errorNotifier: errorNotifier,
      errorViewBuilder: errorViewBuilder,
    );
    return padding != null
        ? SliverPadding(
            padding: padding,
            sliver: sliver,
          )
        : sliver;
  }

  /// Creates a Pagewise SliverList.
  ///
  /// All the properties are either those documented for normal [SliverList](https://docs.flutter.io/flutter/widgets/SliverList-class.html)
  /// or those inherited from [Pagewise]
  PaginatedSliverList({
    super.key,
    bool addSemanticIndexes = true,
    bool addAutomaticKeepAlives = true,
    bool addRepaintBoundaries = true,
    SemanticIndexCallback? semanticIndexCallback,
    int semanticIndexOffset = 0,
    super.pageLoadController,
    super.pageSize,
    super.pageFuture,
    LoadingBuilder? loadingBuilder,
    super.retryBuilder,
    super.noItemsFoundBuilder,
    super.showRetry = false,
    required super.itemBuilder,
    super.itemDummyBuilder,
    super.itemDummyCount,
    ErrorBuilder? errorBuilder,
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
  }) : super(
            loadingBuilder:
                loadingBuilder ?? (context) => const RefreshProgressIndicator(),
            errorBuilder: errorBuilder ??
                (context, error) => PaginateController.errorBuilder(
                      context: context,
                      error: error,
                      defaultTextError: defaultTextError,
                      getLocalizedMessageFromException:
                          getLocalizedMessageFromException,
                      getLocalizedMessageFromHttpResultCode:
                          getLocalizedMessageFromHttpResultCode,
                      errorNotifier: errorNotifier,
                      errorViewBuilder: errorViewBuilder,
                      pageLoadController: pageLoadController,
                    ),
            builder: (PagewiseState<T> state, {int? itemDummyCount}) {
              return SliverList(
                delegate: SliverChildBuilderDelegate(state.itemBuilder,
                    addAutomaticKeepAlives: addAutomaticKeepAlives,
                    addRepaintBoundaries: addRepaintBoundaries,
                    addSemanticIndexes: addSemanticIndexes,
                    semanticIndexCallback:
                        semanticIndexCallback ?? kDefaultSemanticIndexCallback,
                    semanticIndexOffset: semanticIndexOffset,
                    childCount: itemDummyCount ?? state.itemCount),
              );
            });
}

class PaginatedSliverGrid<T> extends Pagewise<T> {
  /// Creates a Pagewise SliverGrid with a crossAxisCount.
  ///
  /// All the properties are either those documented for normal [SliverGrid](https://docs.flutter.io/flutter/widgets/SliverGrid-class.html)
  /// or those inherited from [Pagewise]

  static Widget countWithExtras<T>({
    Key? key,
    bool addSemanticIndexes = true,
    bool addAutomaticKeepAlives = true,
    bool addRepaintBoundaries = true,
    EdgeInsetsGeometry? padding,
    SemanticIndexCallback semanticIndexCallback = kDefaultSemanticIndexCallback,
    int semanticIndexOffset = 0,
    required int crossAxisCount,
    double childAspectRatio = 1.0,
    double crossAxisSpacing = 0.0,
    double mainAxisSpacing = 0.0,
    PagewiseLoadController<T>? pageLoadController,
    int? pageSize,
    PageFuture<T>? pageFuture,
    LoadingBuilder? loadingBuilder,
    RetryBuilder? retryBuilder,
    NoItemsFoundBuilder? noItemsFoundBuilder,
    bool showRetry = false,
    required ItemBuilder<T> itemBuilder,
    ItemBuilder<T>? itemDummyBuilder,
    int? itemDummyCount,
    ErrorBuilder? errorBuilder,
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
  }) {
    var sliver = PaginatedSliverGrid<T>.count(
      key: key,
      addSemanticIndexes: addSemanticIndexes,
      addAutomaticKeepAlives: addAutomaticKeepAlives,
      addRepaintBoundaries: addRepaintBoundaries,
      semanticIndexCallback: semanticIndexCallback,
      semanticIndexOffset: semanticIndexOffset,
      crossAxisCount: crossAxisCount,
      childAspectRatio: childAspectRatio,
      crossAxisSpacing: crossAxisSpacing,
      mainAxisSpacing: mainAxisSpacing,
      pageLoadController: pageLoadController,
      pageSize: pageSize,
      pageFuture: pageFuture,
      loadingBuilder: loadingBuilder,
      retryBuilder: retryBuilder,
      noItemsFoundBuilder: noItemsFoundBuilder,
      showRetry: showRetry,
      itemBuilder: itemBuilder,
      itemDummyBuilder: itemDummyBuilder,
      itemDummyCount: itemDummyCount,
      errorBuilder: errorBuilder,
      defaultTextError: defaultTextError,
      getLocalizedMessageFromException: getLocalizedMessageFromException,
      getLocalizedMessageFromHttpResultCode:
          getLocalizedMessageFromHttpResultCode,
      errorNotifier: errorNotifier,
      errorViewBuilder: errorViewBuilder,
    );
    return padding != null
        ? SliverPadding(
            padding: padding,
            sliver: sliver,
          )
        : sliver;
  }

  PaginatedSliverGrid.count({
    super.key,
    bool addSemanticIndexes = true,
    bool addAutomaticKeepAlives = true,
    bool addRepaintBoundaries = true,
    SemanticIndexCallback? semanticIndexCallback,
    int semanticIndexOffset = 0,
    required int crossAxisCount,
    double childAspectRatio = 1.0,
    double crossAxisSpacing = 0.0,
    double mainAxisSpacing = 0.0,
    super.pageLoadController,
    super.pageSize,
    super.pageFuture,
    LoadingBuilder? loadingBuilder,
    super.retryBuilder,
    super.noItemsFoundBuilder,
    super.showRetry = false,
    required super.itemBuilder,
    super.itemDummyBuilder,
    super.itemDummyCount,
    ErrorBuilder? errorBuilder,
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
  }) : super(
            loadingBuilder:
                loadingBuilder ?? (context) => const RefreshProgressIndicator(),
            errorBuilder: errorBuilder ??
                (context, error) => PaginateController.errorBuilder(
                      context: context,
                      error: error,
                      defaultTextError: defaultTextError,
                      getLocalizedMessageFromException:
                          getLocalizedMessageFromException,
                      getLocalizedMessageFromHttpResultCode:
                          getLocalizedMessageFromHttpResultCode,
                      errorNotifier: errorNotifier,
                      errorViewBuilder: errorViewBuilder,
                      pageLoadController: pageLoadController,
                    ),
            builder: (PagewiseState<T> state, {int? itemDummyCount}) {
              return SliverGrid(
                delegate: SliverChildBuilderDelegate(state.itemBuilder,
                    addAutomaticKeepAlives: addAutomaticKeepAlives,
                    addRepaintBoundaries: addRepaintBoundaries,
                    addSemanticIndexes: addSemanticIndexes,
                    semanticIndexCallback:
                        semanticIndexCallback ?? kDefaultSemanticIndexCallback,
                    semanticIndexOffset: semanticIndexOffset,
                    childCount: itemDummyCount ?? state.itemCount),
                gridDelegate:
                    SliverGridDelegateWithFixedCrossAxisCountAndLoading(
                        crossAxisCount: crossAxisCount,
                        childAspectRatio: childAspectRatio,
                        crossAxisSpacing: crossAxisSpacing,
                        mainAxisSpacing: mainAxisSpacing,
                        itemCount: itemDummyCount ?? state.itemCount),
              );
            });

  static Widget extentWithExtras<T>({
    Key? key,
    bool addSemanticIndexes = true,
    bool addAutomaticKeepAlives = true,
    bool addRepaintBoundaries = true,
    EdgeInsetsGeometry? padding,
    SemanticIndexCallback semanticIndexCallback = kDefaultSemanticIndexCallback,
    int semanticIndexOffset = 0,
    required double maxCrossAxisExtent,
    double childAspectRatio = 1.0,
    double crossAxisSpacing = 0.0,
    double mainAxisSpacing = 0.0,
    PagewiseLoadController<T>? pageLoadController,
    int? pageSize,
    PageFuture<T>? pageFuture,
    LoadingBuilder? loadingBuilder,
    RetryBuilder? retryBuilder,
    NoItemsFoundBuilder? noItemsFoundBuilder,
    bool showRetry = false,
    required ItemBuilder<T> itemBuilder,
    ItemBuilder<T>? itemDummyBuilder,
    int? itemDummyCount,
    ErrorBuilder? errorBuilder,
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
  }) {
    var sliver = PaginatedSliverGrid<T>.extent(
      key: key,
      addSemanticIndexes: addSemanticIndexes,
      addAutomaticKeepAlives: addAutomaticKeepAlives,
      addRepaintBoundaries: addRepaintBoundaries,
      semanticIndexCallback: semanticIndexCallback,
      semanticIndexOffset: semanticIndexOffset,
      maxCrossAxisExtent: maxCrossAxisExtent,
      childAspectRatio: childAspectRatio,
      crossAxisSpacing: crossAxisSpacing,
      mainAxisSpacing: mainAxisSpacing,
      pageLoadController: pageLoadController,
      pageSize: pageSize,
      pageFuture: pageFuture,
      loadingBuilder: loadingBuilder,
      retryBuilder: retryBuilder,
      noItemsFoundBuilder: noItemsFoundBuilder,
      showRetry: showRetry,
      itemBuilder: itemBuilder,
      itemDummyBuilder: itemDummyBuilder,
      itemDummyCount: itemDummyCount,
      errorBuilder: errorBuilder,
      defaultTextError: defaultTextError,
      getLocalizedMessageFromException: getLocalizedMessageFromException,
      getLocalizedMessageFromHttpResultCode:
          getLocalizedMessageFromHttpResultCode,
      errorNotifier: errorNotifier,
      errorViewBuilder: errorViewBuilder,
    );
    return padding != null
        ? SliverPadding(
            padding: padding,
            sliver: sliver,
          )
        : sliver;
  }

  /// Creates a Pagewise SliverGrid with a maxCrossAxisExtent.
  ///
  /// All the properties are either those documented for normal [SliverGrid](https://docs.flutter.io/flutter/widgets/SliverGrid-class.html)
  /// or those inherited from [Pagewise]
  PaginatedSliverGrid.extent({
    super.key,
    bool addSemanticIndexes = true,
    bool addAutomaticKeepAlives = true,
    bool addRepaintBoundaries = true,
    SemanticIndexCallback? semanticIndexCallback,
    int semanticIndexOffset = 0,
    required double maxCrossAxisExtent,
    double childAspectRatio = 1.0,
    double crossAxisSpacing = 0.0,
    double mainAxisSpacing = 0.0,
    super.pageLoadController,
    super.pageSize,
    super.pageFuture,
    super.loadingBuilder,
    super.retryBuilder,
    super.noItemsFoundBuilder,
    super.showRetry = false,
    required super.itemBuilder,
    super.itemDummyBuilder,
    super.itemDummyCount,
    ErrorBuilder? errorBuilder,
    String defaultTextError = "Unknown error",
    LocalizedMessageFromException? getLocalizedMessageFromException,
    LocalizedMessageFromHttpResultCode? getLocalizedMessageFromHttpResultCode,
    ErrorNotifier? errorNotifier,
    ErrorViewBuilder? errorViewBuilder,
  }) : super(
            errorBuilder: errorBuilder ??
                (context, error) => PaginateController.errorBuilder(
                      context: context,
                      error: error,
                      defaultTextError: defaultTextError,
                      getLocalizedMessageFromException:
                          getLocalizedMessageFromException,
                      getLocalizedMessageFromHttpResultCode:
                          getLocalizedMessageFromHttpResultCode,
                      errorNotifier: errorNotifier,
                      errorViewBuilder: errorViewBuilder,
                      pageLoadController: pageLoadController,
                    ),
            builder: (PagewiseState<T> state, {int? itemDummyCount}) {
              return SliverGrid(
                delegate: SliverChildBuilderDelegate(state.itemBuilder,
                    addAutomaticKeepAlives: addAutomaticKeepAlives,
                    addRepaintBoundaries: addRepaintBoundaries,
                    addSemanticIndexes: addSemanticIndexes,
                    semanticIndexCallback:
                        semanticIndexCallback ?? kDefaultSemanticIndexCallback,
                    semanticIndexOffset: semanticIndexOffset,
                    childCount: itemDummyCount ?? state.itemCount),
                gridDelegate:
                    SliverGridDelegateWithMaxCrossAxisExtentAndLoading(
                        maxCrossAxisExtent: maxCrossAxisExtent,
                        childAspectRatio: childAspectRatio,
                        crossAxisSpacing: crossAxisSpacing,
                        mainAxisSpacing: mainAxisSpacing,
                        itemCount: itemDummyCount ?? state.itemCount),
              );
            });
}
